module Jarl
  DEFAULT_JARL_CONFIG_PATH = '~/.jarl'

  #
  # Load config and application definitions
  #
  def self.load(options)
    load_config(options['config'] || DEFAULT_JARL_CONFIG_PATH)
    # puts "Jarl::CLI.options: #{options}"
    abort 'Docker is not installed' unless Docker.installed?
    if Jarl.jarl_files.empty?
      abort 'No *.jarl files found'
    elsif Jarl.applications.empty?
      abort 'No application definitions found'
    end
    true
  end

  #
  # Returns current config
  #
  def self.config
    @config
  end

  #
  # Returns list of found *.jarl files
  #
  def self.jarl_files
    @jarl_files ||= load_jarl_files
  end

  # Returns list of registered applications
  #
  def self.applications
    @applications ||= jarl_files.map do |jf|
      begin
        jf.applications
      rescue StandardError => e
        abort "Failed to parse application definition in file '#{jf.path}': #{e}"
      end
    end.flatten
  end

  # Returns list of running application instances
  #
  def self.application_instances
    @application_instances ||= applications.map(&:instances).flatten
  end

  # Returns list of applications matching the pattern
  #
  # @param patterns [nil] Match all application instances
  # @param patterns [String] Match application instances by single pattern
  # @param patterns [Arary<String>] Match application instances by all patterns
  #
  def self.find_applications_by(patterns)
    find_application_instances_by(patterns).map(&:application).uniq
  end

  # Returns list of running application instances matching the pattern(s)
  #
  # @param patterns [nil] Match all application instances
  # @param patterns [String] Match application instances by single pattern
  # @param patterns [Arary<String>] Match application instances by all patterns
  #
  def self.find_application_instances_by(patterns)
    patterns = nil if patterns && patterns.empty?
    case patterns
    when nil
      application_instances
    when String
      application_instances.select { |i| i.full_name.index(patterns) }
    when Array
      application_instances.select { |i| patterns.any? { |p| i.full_name.index(p) } }
    else
      fail "Unexpected #{patterns.class} passed as argument to .find_application_instances_by"
    end
  end

  #
  #
  def self.reload!
    @jarl_files = nil
    @applications = nil
    @application_instances = nil
    Docker.reload!
  end

  private

  def self.load_config(filename)
    @config = Config.new(filename)
  rescue StandardError => e
    raise "Failed to load config file '#{filename}': #{e}"
  end

  def self.load_jarl_files
    config.params['jarl_files'].map { |p| Pathname.new(p).expand_path }.map do |jarl_files_path|
      Pathname.glob(jarl_files_path).map do |p|
        JarlFile.new(p)
      end
    end.flatten
  end
end # module Jarl
